{% raw %}
<b-modal ref="error-modal" hide-footer title="Something went wrong!">
    <div class="d-block text-center">
        <h3>We're sorry ...</h3>
    </div>
    <h5>
        {{ error_status }}
    </h5>
    <p>
        {{ error_message }}
    </p>
    <b-button class="mt-3" variant="outline-danger" @click="hideErrorModal">Close</b-button>
</b-modal>
{% endraw %}