<!DOCTYPE html>

<html lang="en">
<head>
    <title>GBD Benchmark Database</title>
    {% include "header_content.html" %}
</head>
<body>
{% raw %}
<div class="position-relative" id="app">
    <b-container id="main-content">
        <b-container>
            {% endraw %}
            <b-img center src="{{ url_for('static', filename='resources/gbd_logo_small.png') }}" fluid
                   alt="GBD-Logo"></b-img>
            {% raw %}
        </b-container>
        <b-container>
            <footer>
                {% endraw %}
                {% include "footer_content.html" %}
                {% raw %}
            </footer>
        </b-container>
        {% endraw %}
        {% include "error_modal.html" %}
        {% raw %}
        <b-container>
            <div style="margin-top: 10px">
                {% endraw %}
                {% include "query_form.html" %}
                {% raw %}
            </div>
        </b-container>
        <b-container v-if="table.show" style="padding-bottom: 10px">
            {% endraw %}
            {% include "result_table.html" %}
            {% raw %}
        </b-container>
    </b-container>
</div>
{% endraw %}
</body>
<script type="text/javascript" src="{{ url_for('static', filename='js/index.js') }}">
</script>
</html>
